#include <stdio.h>
#include <math.h>

#include "graph.h"

enum {
    PROP_0,
    PROP_NPTS,
    PROP_XV,
    PROP_YV
};


static void       graph_class_init          (HstGraphClass *klass);
static void       graph_init                (HstGraph      *object,
                                             HstGraphClass *klass);
static void       graph_set_property      (GObject       *object,
                                           guint           property_id,
                                           const GValue   *value,
                                           GParamSpec     *pspec);
static void       graph_get_property      (GObject       *object,
                                           guint           property_id,
                                           GValue         *value,
                                           GParamSpec     *pspec);

static void       graph_dispose             (GObject        *object);
static void       graph_finalize            (GObject        *object);
static void       graph_array(HstFunction *f, Tuple *, double xmin, double xmax);
static double     graph_evaluate(HstFunction *f, double x);

static gpointer    parent_class = NULL;

GType hst_graph_get_type (void) {
    static GType object_type = 0;

    if (!object_type) {
        static const GTypeInfo object_info = {
            sizeof (HstGraphClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) graph_class_init,
            NULL,               /* class_finalize */
            NULL,               /* class_data */
            sizeof (HstGraph),
            16,         /* n_preallocs */
            (GInstanceInitFunc) graph_init,
            NULL,               /* value_table */
        };

        object_type = g_type_register_static (HST_TYPE_FUNCTION, "HstGraph", &object_info, 0);
    }

    return object_type;
}

static void graph_class_init (HstGraphClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (class);
    HstFunctionClass *fclass = HST_FUNCTION_CLASS(class);
    parent_class = g_type_class_ref (HST_TYPE_FUNCTION);

    gobject_class->set_property = graph_set_property;
    gobject_class->get_property = graph_get_property;
    gobject_class->dispose = graph_dispose;
    gobject_class->finalize = graph_finalize;

    fclass->eval  = graph_evaluate;
    fclass->array = graph_array;

#if   GLIB_CHECK_VERSION(2,3,0)
    g_object_class_override_property (
        gobject_class,
        PROP_NPTS,
        "nx"
        );
#else
    g_object_class_install_property (
        gobject_class,
        PROP_NPTS,
        g_param_spec_int ("nx", "Number of points",
                          "Number of points",
                          1,999999,10,
                          G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
#endif

    g_object_class_install_property (
        gobject_class,
        PROP_XV,
        g_param_spec_pointer ("xv", "ptr to abscisas",
                              "Pointer to abscisas",
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );

    g_object_class_install_property (
        gobject_class,
        PROP_YV,
        g_param_spec_pointer ("yv", "ptr to valuess",
                              "Pointer to values",
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
}

static void graph_init( HstGraph *h, HstGraphClass *klass) {
    HstFunction *f;
    g_return_if_fail( HST_IS_GRAPH(h) );

    f = HST_FUNCTION(h);
    f->nbin = 0;
    h->xv= 0;
    h->yv= 0;

    h->ymin = h->ymax = 0.;
}

static void
graph_set_property (GObject      *object,
                    guint         property_id,
                    const GValue *value,
                    GParamSpec   *pspec)
{
    HstFunction *f;

    g_return_if_fail( HST_IS_GRAPH(object) );

    f = HST_FUNCTION(object);

    switch (property_id) {
        case PROP_NPTS:
        case PROP_XV:
        case PROP_YV:
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
graph_get_property(
    GObject     *object,
    guint        property_id,
    GValue      *value,
    GParamSpec  *pspec)
{
    HstGraph *h;
    HstFunction *f;
    g_return_if_fail( HST_IS_GRAPH(object) );

    f = HST_FUNCTION(object);
    h = HST_GRAPH(object);
    switch (property_id) {
    case PROP_NPTS:
        g_value_set_int(value,f->nbin);
        break;
    case PROP_XV:
        g_value_set_pointer (value,(void *)h->xv);
        break;
    case PROP_YV:
        g_value_set_pointer (value,(void *)h->yv);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void graph_dispose(GObject        *object) {
    G_OBJECT_CLASS (parent_class)->dispose (object);
}
static void graph_finalize (GObject        *object) {
    HstGraph *g = HST_GRAPH(object);

    g_free(g->xv);
    g_free(g->yv);

    G_OBJECT_CLASS (parent_class)->finalize (object);
}
void
graph_array(HstFunction *f, Tuple *t, double xmin, double xmax)
{
    int i;
    HstGraph     *g   = HST_GRAPH(f);

    t->n = f->nbin;
    t->P = g_new(PointList,t->n);

    for (i=0;i<t->n;i++) {
        t->P[i].x = g->xv[i];
        t->P[i].y = g->yv[i];
    }
    return;
}

double
graph_evaluate(HstFunction *f, double x)
{
    int i;
    HstGraph     *g   = HST_GRAPH(f);
    for (i=0;i<f->nbin && x>g->xv[i];i++);
    if (i==f->nbin)
        i--;
    if (i==0)
        i++;
    return g->yv[i-1];
}

/**
 * hst_graph_new:
 * @n: number of (x,y) pairs
 * @x: array with values of X
 * @y: array with values of Y
 *
 * Creates a new HstGraph
 *
 * Return value: a HstGraph
 **/
HstGraph *
hst_graph_new(int n, double *x, double *y) {
    HstGraph *g;
    g = g_object_newv (HST_TYPE_GRAPH, 0, NULL);
    hst_graph_reset(g,n,x,y);
    return g;
}

/**
 * hst_graph_reset:
 * @n: number of (x,y) pairs
 * @x: array with values of X
 * @y: array with values of Y
 *
 * Changes the number of points and their values to the ones given at
 * the input. Note: the values are copied into the internal buffers.
 **/
void
hst_graph_reset(HstGraph *g, int n, double *x, double *y) {
    HstFunction *f;
    int i;

    if (g->xv)
        g_free(g->xv);
    if (g->yv)
        g_free(g->yv);

    f = HST_FUNCTION(g);

    g->xv = g_new(double, n);
    g->yv = g_new(double, n);

    f->xmin = g->ymin = G_MAXDOUBLE;
    f->xmax = g->ymax =-G_MAXDOUBLE;
    for (i=0;i<n;i++) {
        g->xv[i] = x[i];
        g->yv[i] = y[i];
        f->xmin = MIN(f->xmin, x[i]);
        f->xmax = MAX(f->xmax, x[i]);
        g->ymin = MIN(g->ymin, y[i]);
        g->ymax = MAX(g->ymax, y[i]);
    }
    f->nbin = n;
}

