#include <stdio.h>
#include <math.h>
#include "gtkgauge.h"
#include "axis.h"

typedef struct _GtkGaugePriv GtkGaugePriv;

struct _GtkGaugePriv {
    int ox, oy;
    int rx, ry;
    int cw, ch;
    double angmn;
    double step;
    double sx, ex;
    GdkGC *gc;
    GdkPixmap *pm;
    GdkPixmap *pxmap;
    GdkFont *ticks;
    GdkFont *label;
    GdkFont *upper;
    GdkColor color;
};


static GtkWidgetClass *parent_class;

static void   gauge_class_init (GtkGaugeClass * klass);
static void   gauge_init (GtkGauge * gg);
static void   gauge_realize (GtkWidget * widget);
static void   gauge_unrealize(GtkWidget *w);
static void   gauge_size_allocate (GtkWidget * widget, GtkAllocation * allocation);
static void   gauge_size_request (GtkWidget *, GtkRequisition *);
static void   gauge_update_scale (GtkGauge * gg);
static gint   gauge_expose (GtkWidget * widget, GdkEventExpose * event);
static void   gauge_paint (GtkWidget * widget, GdkRectangle * area);
static int    make_ellipse (gint xx, gint yy, gint w, gint h, GdkPoint **);
static void   gauge_finalize (GObject *);
static double gauge_get_phi (GtkGauge * gg, double);
static void   gauge_set_property        (GObject	 *object,
					 guint           property_id,
					 const GValue   *value,
					 GParamSpec     *pspec);
static void   gauge_get_property        (GObject	 *object,
					 guint           property_id,
					 GValue         *value,
					 GParamSpec     *pspec);

enum {
  PROP_0,
  PROP_XMIN,
  PROP_XMAX
};


GtkType gtk_gauge_get_type (void)
{
    static GtkType gauge_type = 0;

    if (!gauge_type) {
	static const GtkTypeInfo gauge_info = {
	    "GtkGauge",
	    sizeof (GtkGauge),
	    sizeof (GtkGaugeClass),
	    (GtkClassInitFunc) gauge_class_init,
	    (GtkObjectInitFunc) gauge_init,
	    /* reserved_1 */ NULL,
	    /* reserved_2 */ NULL,
	    (GtkClassInitFunc) NULL,
	};

	gauge_type = gtk_type_unique (GTK_TYPE_WIDGET, &gauge_info);
    }

    return gauge_type;
}

static void gauge_class_init (GtkGaugeClass * class)
{
    GObjectClass   *gobject_class;
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;

    parent_class = g_type_class_peek_parent (class);

    gobject_class = G_OBJECT_CLASS (class);
    object_class = GTK_OBJECT_CLASS (class);
    widget_class = GTK_WIDGET_CLASS (class);

    gobject_class->finalize     = gauge_finalize;
    gobject_class->set_property = gauge_set_property;
    gobject_class->get_property = gauge_get_property;

    widget_class->realize = gauge_realize;
    widget_class->unrealize = gauge_unrealize;
    widget_class->size_allocate = gauge_size_allocate;
    widget_class->expose_event = gauge_expose;
    widget_class->size_request = gauge_size_request;

    g_object_class_install_property (
	gobject_class,
	PROP_XMIN,
	g_param_spec_double("xmin","xmin","Minimum value of X",
			    -DBL_MAX,DBL_MAX,0.,
			    G_PARAM_READABLE | G_PARAM_WRITABLE)
	);
    g_object_class_install_property (
	gobject_class,
	PROP_XMAX,
	g_param_spec_double("xmax","xmax","Maximum value of X",
			    -DBL_MAX,DBL_MAX,1.,
			    G_PARAM_READABLE | G_PARAM_WRITABLE)
	);
}

static void gauge_init (GtkGauge * gg)
{
    gg->val = 0.;
    gg->xmin = 0.;
    gg->xmax = 1.;
    gg->pm = 0;
    gg->pmscale=0;
    gg->pannel = 0.2;
    gg->dot = 0.05;
    gg->title = 0;
    gg->squared = FALSE;
    gtk_gauge_set_color(gg,"red");
}

static void
gauge_set_property (GObject      *object,
		    guint         property_id,
		    const GValue *value,
		    GParamSpec   *pspec)
{
    GtkGauge *ff;
    g_return_if_fail( GTK_IS_GAUGE(object) );

    ff = GTK_GAUGE(object);
    switch (property_id) {
    case PROP_XMIN:
	ff->xmin = g_value_get_double(value);
	break;
    case PROP_XMAX:
	ff->xmax = g_value_get_double(value);
	break;
    default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	break;
    }
}

static void
gauge_get_property (GObject     *object,
		  guint        property_id,
		  GValue      *value,
		  GParamSpec  *pspec)
{
    GtkGauge *ff;
    g_return_if_fail( GTK_IS_GAUGE(object) );

    ff = GTK_GAUGE(object);

    switch (property_id) {
    case PROP_XMIN:
	g_value_set_double(value,ff->xmin);
	break;
    case PROP_XMAX:
	g_value_set_double(value,ff->xmax);
	break;
    default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	break;
    }
}

GtkWidget *gtk_gauge_new (float xmin, float xmax)
{
    GtkGauge *gg = g_object_new ( GTK_TYPE_GAUGE,
				  "xmin",xmin,
				  "xmax", xmax,
				  NULL);

    return GTK_WIDGET (gg);
}

void gauge_finalize (GObject * obj)
{
    GtkGauge *gg;
    g_return_if_fail (obj != NULL);
    g_return_if_fail (GTK_IS_GAUGE (obj));

    gg = GTK_GAUGE (obj);
    if (gg->title) {
	g_free (gg->title);
	gg->title=0;
    }
    if (gg->pm) {
	g_object_unref( gg->pm );
	gg->pm=NULL;
    }
    if (gg->pmscale) {
	g_object_unref( gg->pmscale );
	gg->pmscale=NULL;
    }

    G_OBJECT_CLASS (parent_class)->finalize (obj);
}

static inline double gauge_get_phi (GtkGauge * gg, double x)
{
    double v;
    if ( x>=gg->xmin && x<=gg->xmax )
	v = x;
    else if ( x<gg->xmin) {
	v=gg->xmin;
    }
    else
	v=gg->xmax;

    return gg->angmn - gg->step * (v - gg->xmin);
}



void gtk_gauge_set_value (GtkGauge * gg, float v)
{
    gg->val = v;
    gtk_widget_queue_draw (GTK_WIDGET (gg));
}

float gtk_gauge_get_value (GtkGauge * gg)
{
    return gg->val;
}

void gtk_gauge_set_title (GtkGauge * gg, const char *tit)
{
    if (gg->title)
	g_free (gg->title);
    gg->title = (tit != 0 ? g_strdup (tit) : 0);
    if (GTK_WIDGET_REALIZED (GTK_WIDGET (gg)))
	gtk_widget_queue_resize (GTK_WIDGET (gg));
}

char *gtk_gauge_get_title (GtkGauge * gg)
{
    return gg->title;
}

void gtk_gauge_set_range (GtkGauge * gg, float xmin, float xmax)
{
    gg->xmin = xmin;
    gg->xmax = xmax;
    gauge_update_scale (gg);
    gtk_widget_queue_draw (GTK_WIDGET (gg));
}

void gtk_gauge_set_color (GtkGauge * gg, const char *color_name)
{
    GdkColor col;

    g_return_if_fail( gg!=NULL );
    g_return_if_fail( GTK_IS_GAUGE(gg) );

    if (! gdk_color_parse(color_name,&col) ) {
    	g_printerr("Cannot get color %s\n",color_name);
    	return;
    }
    gg->color = col;

    if (GTK_WIDGET_REALIZED (gg))
	gdk_color_alloc (gtk_widget_get_colormap (GTK_WIDGET (gg)), &gg->color);
}


static void gauge_unrealize(GtkWidget *w) {
    GtkGauge *gg = GTK_GAUGE(w);
    if (gg->pm) {
	g_object_unref( gg->pm );
	gg->pm=NULL;
    }
    if (gg->pmscale) {
	g_object_unref( gg->pmscale );
	gg->pmscale=NULL;
    }
    GTK_WIDGET_CLASS (parent_class)->unrealize(w);
}

static void gauge_realize (GtkWidget * widget)
{
    GtkGauge *gg;
    GdkWindowAttr attributes;
    gint attributes_mask;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_GAUGE (widget));

    gg = GTK_GAUGE (widget);
    GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual (widget);
    attributes.colormap = gtk_widget_get_colormap (widget);
    attributes.event_mask = gtk_widget_get_events (widget) | GDK_EXPOSURE_MASK;

    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

    widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
    gdk_window_set_user_data (widget->window, gg);

    widget->style = gtk_style_attach (widget->style, widget->window);
    gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

    gdk_color_alloc (gtk_widget_get_colormap (widget), &gg->color);
    gg->pm = gdk_pixmap_new (widget->window, widget->allocation.width, widget->allocation.height, -1);
    gauge_update_scale (GTK_GAUGE (widget));

}

static void gauge_size_request (GtkWidget * w, GtkRequisition * requisition)
{
    PangoLayout *layout;
    PangoRectangle logical_rect;

    if (w == NULL)
	return;
    if (!GTK_IS_GAUGE (w))
	return;
    if (requisition == NULL)
	return;

    layout = gtk_widget_create_pango_layout (w, "X");
    pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
    requisition->width  = 15*logical_rect.width;
    requisition->height = requisition->width;
    g_object_unref (layout);
    /* Chain up to default */
    GTK_WIDGET_CLASS (parent_class)->size_request (w, requisition);
}


static void gauge_size_allocate (GtkWidget * widget, GtkAllocation * allocation)
{
    GtkGauge *gg;
    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_GAUGE (widget));
    g_return_if_fail (allocation != NULL);

    /* Chain up to default that simply reads current requisition */
    GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);


    if (GTK_WIDGET_REALIZED (widget)) {
	gg = GTK_GAUGE (widget);
	g_object_unref (gg->pm);
	gg->pm = gdk_pixmap_new (widget->window, widget->allocation.width, widget->allocation.height, -1);
	gauge_update_scale (gg);
    }
}


static void gauge_paint (GtkWidget * widget, GdkRectangle * area)
{
    static char line[50];
    GtkGauge *gg;
    GdkGC *gc;
    double xp;
    int xx, yy;
    PangoLayout *layout;
    PangoRectangle logical_rect;

    g_return_if_fail (widget != NULL);
    g_return_if_fail (GTK_IS_GAUGE (widget));
    gg = GTK_GAUGE (widget);


    gc = gdk_gc_new (gg->pmscale);
    gdk_draw_pixmap (gg->pm, gc, gg->pmscale, 0, 0, 0, 0, -1, -1);

    xp = gauge_get_phi (gg, gg->val);
    xx =  (int) (gg->rx * cos (xp) + .5);
    yy = -(int) (gg->ry * sin (xp) + .5);
    gdk_gc_set_line_attributes (gc, 2, GDK_LINE_SOLID, GDK_CAP_ROUND, GDK_JOIN_ROUND);

    gdk_draw_line (gg->pm, gc, gg->ox, gg->oy, gg->ox + xx, gg->oy + yy);


    sprintf (line, "%.1f", gg->val);
    layout = gtk_widget_create_pango_layout (widget,NULL);
    pango_layout_set_markup(layout,line,-1);
    pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
    gdk_draw_layout (gg->pm, gc, gg->ox-logical_rect.width/2,gg->oy+logical_rect.height/2.,layout);

    gdk_draw_pixmap (widget->window, gc, gg->pm, 0, 0, 0, 0, -1, -1);
    g_object_unref(gc);
    g_object_unref(layout);
}



static gint gauge_expose (GtkWidget * widget, GdkEventExpose * event)
{
    g_return_val_if_fail (widget != NULL, FALSE);
    g_return_val_if_fail (GTK_IS_GAUGE (widget), FALSE);
    g_return_val_if_fail (event != NULL, FALSE);

    if (GTK_WIDGET_DRAWABLE (widget))
	gauge_paint (widget, &event->area);

    return FALSE;
}


static void gauge_update_scale (GtkGauge * gg)
{
    GtkWidget *w;
    int i, cx, cy, ox, oy, rx,ry, npts[3];
    int Ndiv[2] = { 10, 5 }, nBins[2];
    float Wmn[2]={0.,0.}, Wmx[2]={0.,0.}, Bw[2] = { 1., 1. };
    double angmn, angRange, pc, pc2, pcr, pcr2,pcd;
    double x, y, xp;
    int xo, yo, xx, yy,ptit;
    double re,sx,ex;

    GdkGC *gc;
    int ch, cw;
    double cang, cdiag, eps = 0.00001;
    int  x_offset,y_offset;
    static char line[50];
    GdkRegion *rc, *os, *is, *ms, *hRgn, *hRgnm;
    GdkPoint pts[3], *ellipse[3];


    PangoLayout    *layout;
    PangoRectangle  fsize;

    g_return_if_fail (gg != NULL);
    g_return_if_fail (GTK_IS_GAUGE (gg));
    w = GTK_WIDGET (gg);

    /* Get the size of a character */
    layout = gtk_widget_create_pango_layout (w,"X");
    pango_layout_get_pixel_extents (layout, NULL, &fsize);
    ch    = fsize.height;
    cw    = fsize.width;
    cang  = atan2 (ch, cw);
    cdiag = sqrt (ch * ch + cw * cw);

    /* Get dimensions and center of gauge */
    cx = w->allocation.width;
    cy = w->allocation.height;
    x_offset = y_offset = 0;
    if ( gg->squared ) {
	if (cx>cy) {
	    x_offset = (cx-cy)/2;
	    cx = cy;
	}
	else {
	    y_offset = (cy-cx)/2;
	    cy = cx;
	}
    }
    ox = x_offset + cx / 2;
    oy = y_offset + cy / 2;
    ry = rx = MIN(cx/2,cy/2);


    /* Get the order of magnitude of the range */
    get_order_of_magnitude (gg->xmax, 3., &sx, &ex);
    sx = (ex!=0. ? pow (10., ex) : 1.);


    /* Get the binning */
    getbins (gg->xmin, gg->xmax, Ndiv[0], Wmn, Wmx, nBins, Bw);
    if ((gg->xmin - Wmn[0]) > eps) {
	Wmn[0] += Bw[0];
	nBins[0]--;
    }
    if ((Wmx[1] - gg->xmax) > eps) {
	Wmx[0] -= Bw[0];
	nBins[0]--;
    }
    getbins (0, Bw[0], Ndiv[1], Wmn + 1, Wmx + 1, nBins + 1, Bw + 1);

    /*
      Get minumun angle and angle range for the scale.
      We only draw the panel from the diagonal of the center to the
      lower left corner until the diagonal of the center to the lower
      right corner.
     */
    angmn    = atan2 ((double) rx, (double) ry);
    angRange = 2. * (M_PI - fabs (angmn));
    angmn    = atan2 (-(double) ry, -(double) rx);
    if (angmn < 0)
	angmn += 2 * M_PI;

    pc        = gg->pannel;
    pc2       = pc / 2;
    pcr       = (1.-pc);
    pcr2      = (1.-pc2);
    pcd       = (gg->dot);
    gg->ox    = ox;
    gg->oy    = oy;
    gg->rx    = rx;
    gg->ry    = ry;
    gg->angmn = angmn;
    gg->step  = angRange / (gg->xmax - gg->xmin);

    /*
       Build the region that will limit the angular extent of the
       scale.
              (0,0)               (cx,0)
                   +--------------+
		   | 	       	  |
		   | 	       	  |
		   | 	(ox,oy)	  |
		   | 	  x    	  |
		   |    .   .	  |
		   |  .	      .	  |
		   |.  	       	. |
		   +--------------+
	       (0,cy)              (cx,cy)

    */
    pts[0].x = ox-rx;
    pts[0].y = oy+ry;
    pts[1].x = ox;
    pts[1].y = oy;
    pts[2].x = ox+rx;
    pts[2].y = oy+ry;
    rc = gdk_region_polygon (pts, 3, GDK_EVEN_ODD_RULE);

    /*
       Create an elliptical region circunscribed to the pixmap
    */
    npts[0] = make_ellipse (ox-rx, oy-ry, 2*rx, 2*ry, &ellipse[0]);
    os = gdk_region_polygon (ellipse[0], npts[0], GDK_EVEN_ODD_RULE);

    /* outer elipse delimiting the pannel*/
    npts[1] = make_ellipse (ox - pcr*rx , oy -pcr*ry , 2 * pcr * rx, 2 * pcr * ry, ellipse + 1);
    is = gdk_region_polygon (ellipse[1], npts[1], GDK_EVEN_ODD_RULE);

    /* inner elipse of the pannel */
    npts[2] = make_ellipse (ox-pcr2*rx , oy-pcr2*ry ,  2 * pcr2 * rx, 2 * pcr2 * ry, ellipse + 2);
    ms = gdk_region_polygon (ellipse[2], npts[2], GDK_EVEN_ODD_RULE);

    hRgn = gdk_region_copy(os);
    gdk_region_subtract (hRgn, is);
    gdk_region_subtract (hRgn, rc);

    hRgnm = gdk_region_copy(ms);
    gdk_region_subtract (hRgnm, is);
    gdk_region_subtract (hRgnm, rc);


    /* Create the pixmap */
    if (gg->pmscale)
	g_object_unref (gg->pmscale);

    gg->pmscale = gdk_pixmap_new (w->window,
				  w->allocation.width,
				  w->allocation.height,
				  -1);
    gtk_paint_flat_box (w->style, gg->pmscale,
			GTK_STATE_NORMAL,
			GTK_SHADOW_NONE, NULL, w, "",
			0, 0,w->allocation.width, w->allocation.height);


    /* scale */
    gc = gdk_gc_new (gg->pmscale);
    gdk_gc_set_foreground (gc, &gg->color);
    gdk_gc_set_clip_region (gc, hRgn);
    gdk_draw_rectangle (gg->pmscale, gc, TRUE, ox-rx, oy-ry, 2*rx, 2*ry);
    gdk_gc_set_foreground (gc, &w->style->fg[GTK_STATE_NORMAL]);

    /* draw main scale */
    re = (float) sqrt (rx * rx + ry * ry);

    for (x = Wmn[0]; x <= Wmx[0]; x += Bw[0]) {
	xp = gauge_get_phi (gg, x);
	xx =  (int) (rx * cos (xp) + .5);
	yy =  -(int) (ry * sin (xp) + .5);
	gdk_draw_line (gg->pmscale, gc, ox, oy, ox + xx, oy + yy);
    }

    /* draw the circle in the middle */
    gdk_gc_set_clip_region (gc, NULL);


    gdk_draw_arc (gg->pmscale, gc, TRUE,
		  ox - pcd * rx, oy - pcd * ry,
		  2. * pcd * rx, 2. * pcd * ry,
		  0, 360 * 64);


    /* draw tick labels */
    for (x = Wmn[0]; x <= Wmx[0]; x += Bw[0]) {
	xp = gauge_get_phi (gg, x);
	sprintf (line, "%.1f", x / sx);
	pango_layout_set_text(layout,line,-1);
	pango_layout_get_pixel_extents (layout, NULL, &fsize);
	xo = (int) ( (pcr * rx - cdiag) * cos (xp) - fsize.width/2. );
	yo = (int) ( (pcr * ry - cdiag) * sin (xp) + fsize.height/2. );
	gdk_draw_layout (gg->pmscale, gc, ox+xo, oy-yo, layout);
    }

    /* draw the intermadiate ticks */
    gdk_gc_set_clip_region (gc, hRgnm);
    for (x = Wmn[0]; x < Wmx[0]; x += Bw[0]) {
	for (y = Wmn[1]; y < Wmx[1]; y += Bw[1]) {
	    if (x == y)
		continue;
	    xp = gauge_get_phi (gg, x + y);
	    xx = (int) (rx * cos (xp) + .5);
	    yy = -(int) (ry * sin (xp) + .5);
	    gdk_draw_line (gg->pmscale, gc, ox, oy, ox + xx, oy + yy);
	}
    }


    gdk_gc_set_clip_region (gc, NULL);


    /*
       Draw the title
    */
    ptit = cy;
    if (gg->title) {
	sprintf (line, "<b>%s</b>",gg->title);
	pango_layout_set_markup(layout,line,-1);
	pango_layout_get_pixel_extents (layout, NULL, &fsize);
	ptit -= 2+fsize.height;
	gdk_draw_layout (gg->pmscale, gc, gg->ox - fsize.width/2, ptit,layout);
    }
    /*
       Draw the order of magnitude
    */
    if (sx != 1.) {
	sprintf (line, "x10<sup>%d</sup>",(int)ex);
	pango_layout_set_markup(layout,line,-1);
	pango_layout_get_pixel_extents (layout, NULL, &fsize);
	ptit -= 2+fsize.height;
	gdk_draw_layout (gg->pmscale, gc,ox - fsize.width / 2, ptit,layout);
    }




    for (i = 0; i < 3; i++)
	g_free (ellipse[i]);

    gdk_region_destroy (rc);
    gdk_region_destroy (os);
    gdk_region_destroy (is);
    gdk_region_destroy (ms);
    gdk_region_destroy (hRgn);
    gdk_region_destroy (hRgnm);

    g_object_unref(layout);
    g_object_unref(gc);
}

static int make_ellipse (gint xx, gint yy, gint w, gint h, GdkPoint ** ptlst)
{
    int s, x, y, i;
    int *px, *py;
    double d1, d2, a2, b2;
    GdkPoint *pt;
    if (w <= 0 || h <= 0) {
	if (w == 0 || h == 0) {
	    return 0;
	}
	if (w < 0) {		/* negative width */
	    w = -w;
	    xx -= w;
	}
	if (h < 0) {		/* negative height */
	    h = -h;
	    yy -= h;
	}
    }
    s = (w + h + 2) / 2;	/* max size of x,y array */
    px = (int *) g_malloc (s * sizeof (int));	/* 1/4th of ellipse */
    py = (int *) g_malloc (s * sizeof (int));
    i = 0;
    a2 = (w / 2) * (w / 2);
    b2 = (h / 2) * (h / 2);
    x = 0;
    y = (int) (h / 2.);
    d1 = b2 - a2 * (h / 2) + 0.25 * a2;
    px[i] = x;
    py[i] = y;
    i++;
    while (a2 * (y - 0.5) > b2 * (x + 0.5)) {	/* region 1 */
	if (d1 < 0) {
	    d1 = d1 + b2 * (3.0 + 2 * x);
	    x++;
	} else {
	    d1 = d1 + b2 * (3.0 + 2 * x) + 2.0 * a2 * (1 - y);
	    x++;
	    y--;
	}
	px[i] = x;
	py[i] = y;
	i++;
    }
    d2 = b2 * (x + 0.5) * (x + 0.5) + a2 * (y - 1) * (y - 1) - a2 * b2;
    while (y > 0) {		/* region 2 */
	if (d2 < 0) {
	    d2 = d2 + 2.0 * b2 * (x + 1) + a2 * (3 - 2 * y);
	    x++;
	    y--;
	} else {
	    d2 = d2 + a2 * (3 - 2 * y);
	    y--;
	}
	px[i] = x;
	py[i] = y;
	i++;
    }
    s = i;
    *ptlst = pt = (GdkPoint *) g_malloc (4 * s * sizeof (GdkPoint));
    xx += w / 2;
    yy += h / 2;
    for (i = 0; i < s; i++) {	/* mirror */
	x = px[i];
	y = py[i];
	pt[s - i - 1].x = xx + x;
	pt[s - i - 1].y = yy - y;
	pt[s + i].x = xx - x;
	pt[s + i].y = yy - y;
	pt[3 * s - i - 1].x = xx - x;
	pt[3 * s - i - 1].y = yy + y;
	pt[3 * s + i].x = xx + x;
	pt[3 * s + i].y = yy + y;
    }
    g_free (px);
    g_free (py);
    return 4 * s;
}

void
gtk_gauge_set_squared(GtkGauge *gg, gboolean squared)
{
    g_return_if_fail (gg != NULL);
    g_return_if_fail (GTK_IS_GAUGE (gg));

    gg->squared = squared;
    gtk_widget_queue_draw(GTK_WIDGET(gg));
}
