#include <math.h>
#include <gtk/gtk.h>
#include "gtkhistogram2d.h"
#include "gtkhistogram-protected.h"

/* here are local prototypes */
static void hst2d_init (GtkHst2D * o);
static void hst2d_class_init (GtkHst2DClass * c);
static void hst2d_pixmap (GtkHst1D * h);
static void hst2d_frame (GtkHst1D * h);
static void hst2d_realize(GtkWidget *);
static void hst2d_unrealize(GtkWidget *);
static void hst2d_menu(GtkHst1D *h, GtkWidget *w);
static void hst2d_set_property        (GObject   *object,
                                       guint           property_id,
                                       const GValue   *value,
                                       GParamSpec     *pspec);
static void hst2d_get_property        (GObject   *object,
                                       guint           property_id,
                                       GValue         *value,
                                       GParamSpec     *pspec);
static void hst2d_set_2d_colors(GtkHst2D *h, GtkHstColorScheme);
static void hst2d_clear_2d_colors(GtkHst2D *h);
static void hst2d_finalize(GObject *);
static gboolean hst2d_button_press(GtkHst2D *, GdkEventButton *, gpointer *);
static gboolean hst2d_button_release(GtkHst2D *, GdkEventButton *, gpointer *);
static gboolean hst2d_motion_notify(GtkHst2D *, GdkEventMotion *, gpointer *);

/* pointer to the class of our parent */
static GtkHst1DClass *parent_class = NULL;

enum {
    PROP_0,
    PROP_CTYPE,
    PROP_VMAX,
    PROP_VMIN,

};

typedef struct _Zrange Zrange;
struct _Zrange {
    double fmin;
    double fmax;
};

/*
       Returns the color factor for a 2D histogram
*/
#define MxCOL 512
static gushort color_factor(double v, double offs)
{
    double x = v+offs;
    int y = ((int)x) % 6;
    double val=0.;
    switch (y)
    {
    case 0:
        val= x-floor(x);
        break;
    case 1:
    case 2:
        val= 1.;
        break;
    case 3:
        val= 1.-x+floor(x);
        break;
    case 4:
    case 5:
        val =0.;
        break;
    }
    return (gushort)(val*0xffff);
}

GType gtk_hst_color_scheme_get_type (void) {
    static GType etype = 0;
    if (etype == 0) {
        static const GEnumValue values[] = {
            {HST_COLOR_2D_BW,"HST_COLOR_2D_BW","bw"},
            {HST_COLOR_2D_WB,"HST_COLOR_2D_WB","wb"},
            {HST_COLOR_2D_RED,"HST_COLOR_2D_RED","red"},
            {HST_COLOR_2D_GREEN,"HST_COLOR_2D_GREEN","green"},
            {HST_COLOR_2D_BLUE,"HST_COLOR_2D_BLUE","blue"},
            {HST_COLOR_2D_THERMAL,"HST_COLOR_2D_THERMAL","thermal"},
            {HST_COLOR_2D_LAST,"HST_COLOR_2D_LAST","last"},
            { 0, NULL, NULL }
        };
        etype = g_enum_register_static ("GtkHst2DColorScheme", values);
    }
    return etype;

}

GType
gtk_hst2d_get_type (void)
{
    static GType type = 0;

    if (type == 0)
    {
        static const GTypeInfo info =
            {
                sizeof(GtkHst2DClass),
                (GBaseInitFunc) NULL,
                (GBaseFinalizeFunc) NULL,
                (GClassInitFunc) hst2d_class_init,
                (GClassFinalizeFunc) NULL,
                NULL /* class_data */,
                sizeof(GtkHst2D),
                0 /* n_preallocs */,
                (GInstanceInitFunc) hst2d_init,
                NULL
            };

        type = g_type_register_static (GTK_TYPE_HST1D, "GtkHst2D", &info, (GTypeFlags)0);
    }

    return type;
}


static void
hst2d_init (GtkHst2D * o)
{
    o->squared=FALSE;
    o->bsiz=-1;
    o->ntot = o->nindx = 0;
    o->dindex = o->dkey = NULL;

    gtk_signal_connect (GTK_OBJECT (o), 
                        "button_press_event",
                        GTK_SIGNAL_FUNC (hst2d_button_press),
                        NULL);
    gtk_signal_connect (GTK_OBJECT (o), 
                        "button_release_event",
                        GTK_SIGNAL_FUNC (hst2d_button_release),
                        NULL);
    gtk_signal_connect (GTK_OBJECT (o), 
                        "motion_notify_event",
                        GTK_SIGNAL_FUNC (hst2d_motion_notify),
                        NULL);

}

static void
hst2d_class_init (GtkHst2DClass * c)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (c);
    GtkHst1DClass *h1d_class = GTK_HST1D_CLASS(c);
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (c);

    parent_class = g_type_class_ref (GTK_TYPE_HST1D);

    gobject_class->set_property = hst2d_set_property;
    gobject_class->get_property = hst2d_get_property;
    gobject_class->finalize = hst2d_finalize;

    widget_class->realize = hst2d_realize;
    widget_class->unrealize = hst2d_unrealize;
    h1d_class->draw_pixmap = hst2d_pixmap;
    h1d_class->draw_frame = hst2d_frame;
    h1d_class->xtra_menu = hst2d_menu;

    g_object_class_install_property(
        gobject_class, 
        PROP_CTYPE, 
        g_param_spec_int("ctype", "ctype", "Color scheme", 
                         1, 999999, 10,
                         G_PARAM_READABLE | G_PARAM_WRITABLE )
    );

#if   GLIB_CHECK_VERSION(2,3,0)
    g_object_class_override_property (
        gobject_class,
        PROP_VMIN,
        "vmin"
        );
    g_object_class_override_property (
        gobject_class,
        PROP_VMAX,
        "vmax"
        );
#else
    g_object_class_install_property (
        gobject_class,
        PROP_VMIN,
        g_param_spec_double("vmin",
                            "vmin",
                            "Minimum value to be shown",
                            -DBL_MAX, DBL_MAX, 0.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE
            )
        );

    g_object_class_install_property (
        gobject_class,
        PROP_VMAX,
        g_param_spec_double("vmax",
                            "vmax",
                            "Maximum value to be shown",
                            -DBL_MAX, DBL_MAX, 1.,
                            G_PARAM_READABLE | G_PARAM_WRITABLE
            )
        );
#endif
}

static void
hst2d_set_property (GObject      *object,
                    guint         property_id,
                    const GValue *value,
                    GParamSpec   *pspec)
{
    GtkHst2D *hh;
    GtkHst1D *h1d;
    g_return_if_fail( GTK_IS_HST2D(object) );
    hh = GTK_HST2D(object);
    h1d= GTK_HST1D(object);
    switch (property_id)
    {
    case PROP_CTYPE:
        hh->ctype = (GtkHstColorScheme)g_value_get_int(value);
        break;
    case PROP_VMIN:
        h1d->vmin = g_value_get_double(value);
        break;
    case PROP_VMAX:
        h1d->vmax = g_value_get_double(value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}
static void
hst2d_get_property (GObject     *object,
                    guint        property_id,
                    GValue      *value,
                    GParamSpec  *pspec)
{
    GtkHst2D *hh;
    GtkHst1D *h1d;
    g_return_if_fail( GTK_IS_HST2D(object) );

    hh = GTK_HST2D(object);
    h1d= GTK_HST1D(object);

    switch (property_id)
    {
    case PROP_CTYPE:
        g_value_set_int(value, hh->ctype);
        break;
    case PROP_VMIN:
        g_value_set_double(value, h1d->vmin);
        break;
    case PROP_VMAX:
        g_value_set_double(value, h1d->vmax);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}



GtkWidget *
gtk_hst2d_new(HstHistogram2D *ff)
{
    GtkWidget *w;

    g_return_val_if_fail ( ff!=0, NULL);
    w = g_object_new (GTK_TYPE_HST2D, "func",ff,"ctype",HST_COLOR_2D_THERMAL,NULL);

    return w;
}

static void
hst2d_finalize(GObject *object)
{
    GtkHst2D *h;
    g_return_if_fail( GTK_IS_HST2D(object) );

    h= GTK_HST2D(object);

    hst2d_clear_2d_colors(h);

    g_free(h->dindex);
    h->bsiz = -1;
    h->ntot = h->nindx = 0;
    h->dindex = h->dkey = NULL;

    G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
hst2d_unrealize(GtkWidget *w)
{
    hst2d_clear_2d_colors( GTK_HST2D(w));
    GTK_WIDGET_CLASS (parent_class)->unrealize(w);
}

static void
hst2d_realize(GtkWidget *w)
{
    g_return_if_fail( GTK_IS_HST2D(w) );
    GTK_WIDGET_CLASS (parent_class)->realize (w);
    if ( GTK_WIDGET_REALIZED(w))
    {
        GtkHst2D *h= GTK_HST2D(w);
        hst2d_set_2d_colors(h, h->ctype);
    }
}

#define M 7
#define NSTACK 128
#define swap(a,b) {itmp=(a);(a)=(b);(b)=itmp;}

static void __sort(int n, gdouble *arr, gint *ptr, gint *indx)
{
    static int istack[NSTACK];
    int i, j, k, l=0, ir=n-1, itmp;
    int jstack=0, indxt;
    double a;

    for (i=0; i<n; i++)
        indx[i]=i;

    while (1)
    {
        if (ir-l < M)
        {
            for (j=l+1; j<=ir; j++)
            {
                indxt=indx[j];
                a=arr[ptr[indxt]];
                for (i=j-1; i>=l; i--)
                {
                    if (arr[ptr[indx[i]]] <= a)
                        break;
                    indx[i+1]=indx[i];
                }
                indx[i+1]=indxt;
            }
            if (jstack == 0)
                break;
            ir=istack[jstack--];
            l=istack[jstack--];
        }
        else
        {
            k=(l+ir) >> 1;
            swap(indx[k],indx[l+1]);
            if (arr[ptr[indx[l]]] > arr[ptr[indx[ir]]])
            {
                swap(indx[l],indx[ir]);
            }
            if (arr[ptr[indx[l+1]]] > arr[ptr[indx[ir]]])
            {
                swap(indx[l+1],indx[ir]);
            }
            if (arr[ptr[indx[l]]] > arr[ptr[indx[l+1]]])
            {
                swap(indx[l],indx[l+1]);
            }
            i=l+1;
            j=ir;
            indxt=indx[l+1];
            a=arr[ptr[indxt]];
            for (;;)
            {
                do i++; while (arr[ptr[indx[i]]] < a);
                do j--; while (arr[ptr[indx[j]]] > a);
                if (j < i) break;
                swap(indx[i],indx[j]);
            }
            indx[l+1]=indx[j];
            indx[j]=indxt;
            jstack += 2;
            if (jstack > NSTACK)
            {
                g_printerr("gtkhistogram2d:sort NSTACK too small");
                return;
            }
            if (ir-i+1 >= j-l)
            {
                istack[jstack]=ir;
                istack[jstack-1]=i;
                ir=j-1;
            }
            else
            {
                istack[jstack]=j-1;
                istack[jstack-1]=l;
                l=i;
            }
        }
    }
}

static void
hst2d_pixmap(GtkHst1D * h)
{
    int i, j, ix, iy;
    gdouble cstep, val, delta, sx, sy;
    GdkRectangle cscale;
    GdkRectangle area;
    GtkHst2D *h2d;
    HstHistogram2D *f2d;
    GdkGC *gc;
    Zrange *zr;

    g_return_if_fail (h != NULL);
    g_return_if_fail (GTK_IS_HST2D (h));

    h2d = GTK_HST2D(h);
    if ( h2d->ncol==0 )
        return;

    f2d = HST_HISTOGRAM2D(h->func);
    gc = gdk_gc_new( h->pm );

    /* check if there is a range defined  and draw it */
    zr = g_object_get_data(G_OBJECT(h),"zrange");
    if (!zr)
        zr = g_object_get_data(G_OBJECT(h),"__zrange");

    if (zr)
    {
        double fmin = (zr->fmin<0. ? 0. : (zr->fmin>1. ? 1. : zr->fmin));
        double fmax = (zr->fmax<0. ? 0. : (zr->fmax>1. ? 1. : zr->fmax));

        int pmin = h->cnvs->pz0 - (int)(MAX(fmin,fmax)*h->cnvs->wz);
        int pmax = h->cnvs->pz0 - (int)(MIN(fmin,fmax)*h->cnvs->wz);

        gdk_draw_rectangle(h->pm,gc,TRUE,
                           h->cnvs->cscale.x + h->cnvs->cscale.width + 5,
                           pmin,
                           2,pmax-pmin
            );
    }

    /* draw the scale */
    cstep = h->cnvs->cscale.height/((float)h2d->ncol);
    area.x = h->cnvs->cscale.x;
    area.y = h->cnvs->cscale.y;
    area.width = h->cnvs->cscale.width;
    area.height = MAX(1,cstep+1.);
    val = (double)area.y;
    for (i=h2d->ncol-1;i>=0;i--)
    {
        val += cstep;
        gdk_gc_set_foreground(gc,h2d->colorlst+i);
        gdk_gc_set_clip_rectangle(gc,&area);
        gdk_draw_rectangle (h->pm, gc, TRUE,
                            area.x,
                            area.y,
                            area.width,
                            area.height);
        area.y = val;
    }

    /* draw the contents */
    if (h2d->nindx==0)
    {
        g_object_unref(gc);
        return;
    }

    delta = h->cnvs->zmax - h->cnvs->zmin;
    if (delta==0.)
    {
        g_object_unref(gc);
        return;
    }
    cstep = (h2d->ncol-1.)/delta;
    sx = HST_FUNCTION2D_X(f2d)->step*h->cnvs->xstp;
    sy = HST_FUNCTION2D_Y(f2d)->step*h->cnvs->ystp;
    area.x = h->cnvs->px0;
    area.width = MAX(1,(int)(sx+1.));
    area.height= MAX(1,(int)(sy+1.));
    for (j=0;j<h2d->nindx;j++)
    {
        int ik = h2d->dindex[h2d->dkey[j]];
        val = f2d->y[ik]-h->cnvs->zmin;

        if ( val<0. )
            continue;
        
        if ( val>delta )
            break;

        hst_function2d_find_cell_from_bin(HST_FUNCTION2D(f2d),ik,&ix,&iy);
        area.x = h->cnvs->px0 + ix*sx;
        area.y = h->cnvs->py0 - (iy+1)*sy;
        i = (int)(cstep*val);
        if (i>=h2d->ncol)
            i=h2d->ncol;
        if ( i>=0 )
        {
            gdk_gc_set_foreground(gc,h2d->colorlst+i);
            gdk_gc_set_clip_rectangle(gc,&area);
            gdk_draw_rectangle (h->pm, gc, TRUE,
                                area.x, area.y,
                                area.width, area.height);
        }
    }
    g_object_unref(gc);
    return;
}

static void
hst2d_frame(GtkHst1D * h)
{
    int i, j, ix, iy;
    HstHistogram2D *f2d;
    GtkHst2D *h2d;
    HstHistogram2D *h2;
    Canvas *cnvs;
    g_return_if_fail (h != NULL);
    g_return_if_fail (GTK_IS_HST2D (h));

    f2d = HST_HISTOGRAM2D(h->func);
    h2d = GTK_HST2D(h);
    cnvs = h->cnvs;

    /*
      We want to draw first the smallest values so that they do not
      get drawn on top of the higher ones, which are the ones we
      prefer to see. So we now get an ordered list of the non empty
      bins and we will draw them after.
    */
    h2d->ntot = HST_FUNCTION2D_X(f2d)->nbin*HST_FUNCTION2D_Y(f2d)->nbin;

    /* Count non empty cells */
    h2d->nindx = 0;

    for (ix=0;ix<h2d->ntot;ix++)
    {
        if (f2d->y[ix]!=0.)
            h2d->nindx++;
    }
    if (h2d->nindx==0)
    {
        cnvs->zmin=0.;
        cnvs->zmax=0.;
    }
    else
    {
        Zrange *zr = g_object_get_data(G_OBJECT(h),"zrange");

        /* allocate and prepare the index*/
        int s = 2*h2d->nindx;
        if ( s > h2d->bsiz )
        {
            h2d->bsiz=s;
            h2d->dindex = g_new(gint, s);
        }
        h2d->dkey = h2d->dindex + h2d->nindx;
        for (ix=0,iy=0;ix<h2d->ntot;ix++)
        {
            if ( f2d->y[ix]!=0. )
                h2d->dindex[iy++]=ix;
        }

        /* Sort */
        __sort(h2d->nindx,f2d->y,h2d->dindex,h2d->dkey);
        cnvs->zmax = f2d->y[h2d->dindex[h2d->dkey[h2d->nindx-1]]];
        cnvs->zmin = f2d->y[h2d->dindex[h2d->dkey[0]]];
        if (zr)
        {
            /*
             * In this case we take the second highest and the second
             * smallest to avoid "spikes" spoiling the color scale
             */
            double top = f2d->y[h2d->dindex[h2d->dkey[h2d->nindx-2]]];
            double offset = f2d->y[h2d->dindex[h2d->dkey[1]]];
            double width = top-offset;
            cnvs->zmax = offset + zr->fmax*width;
            cnvs->zmin = offset + zr->fmin*width;
        }
    }
    
    h->xmin = h->cnvs->xmin = HST_FUNCTION2D_X(f2d)->xmin;
    h->xmax = h->cnvs->xmax = HST_FUNCTION2D_X(f2d)->xmax;

    g_object_set(G_OBJECT(h),
                 "hymin",HST_FUNCTION2D_Y(f2d)->xmin,
                 "hymax",HST_FUNCTION2D_Y(f2d)->xmax,
                 NULL
        );

    h2 = HST_HISTOGRAM2D(h->func);
    if (h2)
    {
        if (!cnvs->fixed)
        {
            g_object_get(G_OBJECT(h),
                         "hymin",&cnvs->ymin,
                         "hymax",&cnvs->ymax,
                         NULL);
        }
    }
    return;
}

static void
hst2d_clear_2d_colors(GtkHst2D *h)
{
    if (h->colorlst)
    {
        gdk_colormap_free_colors (
            gtk_widget_get_colormap (GTK_WIDGET(h)),
            h->colorlst,
            h->ncol);
        g_free(h->colorlst);
        h->colorlst=0;
        h->ncol=0;
    }
}

static void
hst2d_set_2d_colors(GtkHst2D *h,GtkHstColorScheme ityp)
{
    int i;
    int     nincycle=25;
    int     tp_ncycle[HST_COLOR_2D_LAST]={1,1,1,1,1,4};
    double  tp_offs[HST_COLOR_2D_LAST][3] = {
        {3,3,3},
        {0,0,0},
        {3,5,5},
        {5,3,5},
        {5,5,3},
        {4,6,2}
    };
    double  cc,cstep;

    g_return_if_fail (h != NULL);
    g_return_if_fail (GTK_IS_HST2D (h));

    hst2d_clear_2d_colors( h );
    h->ctype = ityp;

    h->ncol=tp_ncycle[h->ctype]*nincycle;
    cc = tp_ncycle[ityp]/((float)h->ncol);
    h->colorlst = g_new(GdkColor,h->ncol);
    for (i=0;i<h->ncol;i++)
    {
        cstep = i*cc;
        h->colorlst[i].pixel = 0;
        h->colorlst[i].red = color_factor(cstep,tp_offs[ityp][0]);
        h->colorlst[i].green = color_factor(cstep,tp_offs[ityp][1]);
        h->colorlst[i].blue = color_factor(cstep,tp_offs[ityp][2]);
        gdk_color_alloc (gtk_widget_get_colormap (GTK_WIDGET(h)),h->colorlst+i);
    }

}

void
gtk_hst2d_set_squared(GtkHst2D *h2d, gboolean squared)
{
    g_return_if_fail (h2d != NULL);
    g_return_if_fail (GTK_IS_HST2D (h2d));

    h2d->squared = squared;
    gtk_widget_queue_draw( GTK_WIDGET(h2d) );
}

static void
hst2d_set_squared(GtkWidget *w,GtkHst2D *h2d)
{
    h2d->squared = gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(w));
    gtk_widget_queue_draw( GTK_WIDGET(h2d) );
}


static void
hst2d_set_bw(GtkWidget *w, GtkHst2D *h)
{
    hst2d_set_2d_colors(h, HST_COLOR_2D_BW);
}

static void
hst2d_set_wb(GtkWidget *w, GtkHst2D *h)
{
    hst2d_set_2d_colors(h, HST_COLOR_2D_WB);
}

static void
hst2d_set_red(GtkWidget *w, GtkHst2D *h)
{
    hst2d_set_2d_colors(h, HST_COLOR_2D_RED);
}

static void
hst2d_set_green(GtkWidget *w, GtkHst2D *h)
{
    hst2d_set_2d_colors(h, HST_COLOR_2D_GREEN);
}

static void
hst2d_set_blue(GtkWidget *w, GtkHst2D *h)
{
    hst2d_set_2d_colors(h, HST_COLOR_2D_BLUE);
}

static void
hst2d_set_thermal(GtkWidget *w, GtkHst2D *h)
{
    hst2d_set_2d_colors(h, HST_COLOR_2D_THERMAL);
}


static void
hst2d_menu(GtkHst1D *hst, GtkWidget *menu)
{
    GtkWidget *item;
    GtkHst2D *h=NULL;
    g_return_if_fail( hst!=NULL && menu!=NULL);
    g_return_if_fail( GTK_IS_MENU(menu) );
    g_return_if_fail( GTK_IS_HST2D(hst) );

    h = GTK_HST2D(hst);

    item = gtk_menu_item_new ();
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_widget_set_sensitive (item, FALSE);

    item = gtk_check_menu_item_new_with_mnemonic ("Squared");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),h->squared);
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_signal_connect (GTK_OBJECT (item), "activate",
                        GTK_SIGNAL_FUNC (hst2d_set_squared),
                        h);

    item = gtk_menu_item_new ();
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_widget_set_sensitive (item, FALSE);

    item = gtk_check_menu_item_new_with_mnemonic ("Black/White");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),h->ctype==HST_COLOR_2D_BW);
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_signal_connect (GTK_OBJECT (item), "activate",
                        GTK_SIGNAL_FUNC (hst2d_set_bw),
                        h);

    item = gtk_check_menu_item_new_with_mnemonic ("White/Black");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),h->ctype==HST_COLOR_2D_WB);
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_signal_connect (GTK_OBJECT (item), "activate",
                        GTK_SIGNAL_FUNC (hst2d_set_wb),
                        h);

    item = gtk_check_menu_item_new_with_mnemonic ("Red");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),h->ctype==HST_COLOR_2D_RED);
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_signal_connect (GTK_OBJECT (item), "activate",
                        GTK_SIGNAL_FUNC (hst2d_set_red),
                        h);

    item = gtk_check_menu_item_new_with_mnemonic ("Green");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),h->ctype==HST_COLOR_2D_GREEN);
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_signal_connect (GTK_OBJECT (item), "activate",
                        GTK_SIGNAL_FUNC (hst2d_set_green),
                        h);

    item = gtk_check_menu_item_new_with_mnemonic ("Blue");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),h->ctype==HST_COLOR_2D_BLUE);
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_signal_connect (GTK_OBJECT (item), "activate",
                        GTK_SIGNAL_FUNC (hst2d_set_blue),
                        h);

    item = gtk_check_menu_item_new_with_mnemonic ("Thermal");
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),h->ctype==HST_COLOR_2D_THERMAL);
    gtk_widget_show (item);
    gtk_container_add (GTK_CONTAINER (menu), item);
    gtk_signal_connect (GTK_OBJECT (item), "activate",
                        GTK_SIGNAL_FUNC (hst2d_set_thermal),
                        h);

}

static gboolean
in_color_scale(GtkHst2D *hst,GdkEventButton *ev)
{
    static const int tolerance=15;
    double  x     = ev->x;
    double  y     = ev->y;
    Canvas *cnvs  = GTK_HST1D(hst)->cnvs;
    double  wd    = cnvs->cscale.width/2.;
    double  xdist = x - cnvs->cscale.x - wd;

    if ( fabs(xdist)>wd )
        return FALSE;

    if (  y < (cnvs->pz1-tolerance) )
        return FALSE;

    if ( y > cnvs->pz0+tolerance )
        return FALSE;

    return TRUE;
}

#define FRACTION(y,cnvs) ((cnvs)->pz0 - (y) )/((double)((cnvs)->pz0-(cnvs)->pz1))

static gboolean
hst2d_button_press(GtkHst2D *hst, GdkEventButton *ev, gpointer *d)
{
    if ( ev->button==2)
    {
        Zrange *zr = g_object_steal_data(G_OBJECT(hst),"zrange");
        if (zr)
            g_free(zr);

        return FALSE;
    }
    if ( ev->button!=1)
        return FALSE;

    if ( in_color_scale(hst,ev) )
    {
        Zrange *zr = g_new0(Zrange,1);
        zr->fmax = zr->fmin = FRACTION(ev->y,GTK_HST1D(hst)->cnvs);
        zr->fmin = (zr->fmin<0. ? 0. : (zr->fmin>1. ? 1. : zr->fmin));
        g_object_set_data(G_OBJECT(hst),"__zrange",zr);
    }
    return FALSE;
}

static gboolean
hst2d_button_release(GtkHst2D *hst, GdkEventButton *ev, gpointer *data)
{
    double xd,x = ev->x;
    double yd,y = ev->y;

    Zrange *zr0,*zr = g_object_steal_data(G_OBJECT(hst),"__zrange");
    if (!zr)
        return;

    /*
     * Find if we are on top of the color scale and, if so,
     * store the new Zrange
     */
    if (in_color_scale(hst,ev))
    {
        zr->fmax = FRACTION(ev->y,GTK_HST1D(hst)->cnvs);
        zr->fmax = (zr->fmax<0. ? 0. : (zr->fmax>1. ? 1. : zr->fmax));
        
        if (zr->fmin>zr->fmax)
        {
            double tmp = zr->fmin;
            zr->fmin = zr->fmax;
            zr->fmax = tmp;
        }
        zr0 = g_object_steal_data(G_OBJECT(hst),"zrange");
        if (zr0)
            g_free(zr0);
        g_object_set_data(G_OBJECT(hst),"zrange",zr);
    }
    else
    {
        if (zr)
            g_free(zr);
    }

    return FALSE;
}

static gboolean
hst2d_motion_notify(GtkHst2D *hst, GdkEventMotion *ev, gpointer *data)
{
    Zrange *zr = g_object_get_data(G_OBJECT(hst),"__zrange");
    if (zr)
        zr->fmax = FRACTION(ev->y,GTK_HST1D(hst)->cnvs);
    return FALSE;
}
