#include <stdio.h>
#include <math.h>
#include "histogram.h"

enum {
    CHANGED,
    LAST_SIGNAL
};
enum {
    PROP_0,
    PROP_PTR,
    PROP_VMIN,
    PROP_VMAX,
    PROP_STATS,
};



static void       hst_class_init          (HstHistogramClass *klass);
static void       hst_init                (HstHistogram      *object,
                                           HstHistogramClass *klass);
static void       hst_set_property        (GObject       *object,
                                           guint           property_id,
                                           const GValue   *value,
                                           GParamSpec     *pspec);
static void       hst_get_property        (GObject       *object,
                                           guint           property_id,
                                           GValue         *value,
                                           GParamSpec     *pspec);

static void       hst_dispose             (GObject        *object);
static void       hst_finalize            (GObject        *object);
static void       hst_array(HstFunction *f, Tuple *, double xmin, double xmax);
static double     hst_evaluate(HstFunction *f, double x);

static void   hst_reset(HstHistogram *hst);
static void   hst_fill(HstHistogram *hst, double x, double w);
static void   hst_real_changed(HstHistogram*hst);

static gpointer    parent_class = NULL;
static guint       hst_signals[LAST_SIGNAL] = { 0 };

GType hst_histogram_get_type (void)
{
    static GType object_type = 0;

    if (!object_type)
    {
        static const GTypeInfo object_info = {
            sizeof (HstHistogramClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) hst_class_init,
            NULL,               /* class_finalize */
            NULL,               /* class_data */
            sizeof (HstHistogram),
            16,         /* n_preallocs */
            (GInstanceInitFunc) hst_init,
            NULL,               /* value_table */
        };

        object_type = g_type_register_static (HST_TYPE_FUNCTION, "HstHistogram", &object_info, 0);
    }

    return object_type;
}

static void
hst_class_init (HstHistogramClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (class);
    HstFunctionClass *fclass = HST_FUNCTION_CLASS(class);
    parent_class = g_type_class_ref (HST_TYPE_FUNCTION);

    gobject_class->set_property = hst_set_property;
    gobject_class->get_property = hst_get_property;
    gobject_class->dispose = hst_dispose;
    gobject_class->finalize = hst_finalize;

    fclass->eval  = hst_evaluate;
    fclass->array = hst_array;

    class->fill  = hst_fill;
    class->reset = hst_reset;
    class->changed = NULL;
    /**
     * HstHistogram:ptr
     * 
     * Sets and retrieves the pointer of the array with the
     * histogram contents.
     */
    g_object_class_install_property (
        gobject_class,
        PROP_PTR,
        g_param_spec_pointer ("ptr", "ptr to data",
                              "Pointer to data",
                              G_PARAM_READABLE | G_PARAM_WRITABLE)
        );
    
    /**
     * HstHistogram:vmin
     * 
     * Minimum value of hte histogram contents.
     * 
     */
    g_object_class_install_property (
        gobject_class,
        PROP_VMIN,
        g_param_spec_double("vmin","vmin","Minimum histogramed value",
                            -DBL_MAX,DBL_MAX,0.,
                            G_PARAM_READABLE)
        );
    
    /**
     * HstHistogram:vmax
     * 
     * Maximum value of the histogram contents
     * 
     */
    g_object_class_install_property (
        gobject_class,
        PROP_VMAX,
        g_param_spec_double("vmax","vmax","Maximum histogramed value",
                            -DBL_MAX,DBL_MAX,1.,
                            G_PARAM_READABLE )
        );

    /**
     * HstHistogram:stats
     *
     * Setting any value on this property will update the statistics
     * of the histogram.
     **/
    g_object_class_install_property (
        gobject_class,
        PROP_STATS,
        g_param_spec_int("stats","stats","Updates the statistics",
                         -INT_MAX,INT_MAX,0,
                         G_PARAM_WRITABLE)
        );
    /**
     * HstHistogram::changed
     *
     * The changed signal is emitted whenever the contents of the
     * histogram are changed via the ::hst_histogram_fill or
     * ::hst_histogram_reset functions.
     **/
    hst_signals[CHANGED]
        = g_signal_new ("changed",
                        G_TYPE_FROM_CLASS (gobject_class),
                        G_SIGNAL_RUN_FIRST,
                        G_STRUCT_OFFSET (HstHistogramClass, changed),
                        NULL, NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE, 0);

}

static void
hst_init( HstHistogram *h, HstHistogramClass *klass)
{
    HstFunction *f;
    g_return_if_fail( HST_IS_HISTOGRAM(h) );

    f = HST_FUNCTION(h);
    h->y = g_new(double,f->nbin);
    h->owner = 1;
    h->stat = hst_stats_new();
    h->vmin = h->vmax = 0.;
}

static void
hst_set_property (GObject      *object,
                  guint         property_id,
                  const GValue *value,
                  GParamSpec   *pspec)
{
    HstHistogram *h;
    HstFunction  *f;
    void      *ptr;
    g_return_if_fail( HST_IS_HISTOGRAM(object) );

    h = HST_HISTOGRAM(object);
    f = HST_FUNCTION(object);
    switch (property_id)
    {
    case PROP_PTR:
        ptr =  g_value_get_pointer (value);
        if (ptr==0)
        {
            if (h->owner)
                g_free(h->y);
            h->y = g_new(double,f->nbin);
            h->owner=1;
        }
        else
        {
            int ix;
            if (h->owner)
                g_free(h->y);
            h->y = (double *)ptr;
            h->owner=0;
            hst_histogram_update_stats(h);
        }
        break;
    case PROP_STATS:
        hst_histogram_update_stats(h);
        break;
        
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void
hst_get_property (GObject     *object,
                  guint        property_id,
                  GValue      *value,
                  GParamSpec  *pspec)
{
    HstHistogram *h;
    g_return_if_fail( HST_IS_HISTOGRAM(object) );

    h = HST_HISTOGRAM(object);

    switch (property_id)
    {
    case PROP_PTR:
        g_value_set_pointer (value,(void *)h->y);
        break;
    case PROP_VMIN:
        g_value_set_double(value,h->vmin);
        break;
    case PROP_VMAX:
        g_value_set_double(value,h->vmax);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

/**
 * hst_histogram_new:
 * @n: number of bins in the histogram
 * @xmin: lower range of the X axis
 * @xmax: upper range of the X axis
 * @ptr: a pointer to an array of doubles. HstHistogram will use it
 *       internally as the contents of the histogram. It can be
 *       used to monitor a given array or to retrieve or fill the
 *       buffer with the contents of the histogram
 *
 * This is a convenience function to create a #HstHistogram and
 * setting its properties in one go.
 *
 * Return value: a HstHistogram pointer
 *
 */
HstHistogram *
hst_histogram_new(int n, double xmin, double xmax, double *ptr)
{
    HstFunction *f;
    HstHistogram *h;
    g_return_val_if_fail ( n>0, NULL );
    h = g_object_new (HST_TYPE_HISTOGRAM,
                      "nx",n,
                      "xmin",xmin,
                      "xmax",xmax,
                      "ptr",ptr,
                      NULL);
    hst_histogram_reset(h);
    return h;
}

static void
hst_dispose(GObject *object)
{
    G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
hst_finalize (GObject *object)
{
    HstHistogram *hst = HST_HISTOGRAM(object);

    if (hst->owner)
    {
        g_free(hst->y);
        hst->y=0;
    }
    g_object_unref(G_OBJECT(hst->stat));


    G_OBJECT_CLASS (parent_class)->finalize (object);
}



static double
hst_evaluate(HstFunction *f, double x)
{
    HstHistogram *hst;
    int wbin;

    g_return_val_if_fail ( HST_IS_HISTOGRAM(f), 0.);
    hst = HST_HISTOGRAM(f);
    if (x<f->xmin || x>f->xmax )
        return 0.;

    wbin = hst_function_find_bin(f,x);
    if (wbin<0)
        wbin=0;
    if (wbin>=f->nbin)
        wbin = f->nbin-1;
    return hst->y[wbin];
}

static void
hst_reset(HstHistogram *hst)
{
    HstFunction *f;
    int i;

    g_return_if_fail( HST_IS_HISTOGRAM(hst) );

    f = HST_FUNCTION(hst);
    for (i=0;i<f->nbin;i++)
        hst->y[i]=0;
    hst_stats_reset(hst->stat);
    hst->vmin = 0.;
    hst->vmax = 0.;
}

static void
hst_fill(HstHistogram *hst, double x, double w)
{
    HstFunction *f;
    int wbin;

    g_return_if_fail( HST_IS_HISTOGRAM(hst) );

    f = HST_FUNCTION(hst);
    wbin = hst_function_find_bin(HST_FUNCTION(hst),x);
    hst_stats_add(hst->stat,x,0.,w);

    if ( wbin<0 || wbin>=f->nbin )
        return;

    hst->y[wbin]+=w;

    if ( hst->y[wbin]> hst->vmax )
        hst->vmax = hst->y[wbin];
    if ( hst->y[wbin] < hst->vmin )
        hst->vmin = hst->y[wbin];
}


/**
 * hst_histogram_mean:
 * @hst: a #HstHistogram
 *
 * Return value: the mean value
 **/
double
hst_histogram_mean(HstHistogram *hst)
{
    return hst_stats_get_mean(hst->stat,0);
}

/**
 * hst_histogram_rms:
 * @hst: a #HstHistogram
 *
 * Return value: the RMS value
 **/
double
hst_histogram_rms(HstHistogram *hst)
{
    return hst_stats_get_rms(hst->stat,0);
}


/**
 * hst_histogram_stdev:
 * @hst: a #HstHistogram
 *
 * Return value: the standar deviation of the distribution
 *               in the histogram.
 **/
double
hst_histogram_stdev(HstHistogram *hst)
{
    return hst_stats_get_stdev(hst->stat,0);
}

/**
 * hst_histogram_n:
 * @hst: a #HstHistogram
 *
 * Return value: the number of entries
 **/
double
hst_histogram_n(HstHistogram *hst)
{
    return hst_stats_get_nc(hst->stat);
}

/**
 * hst_histogram_xmin:
 * @hst: a #HstHistogram
 *
 * Return value: the lower value of the X axis
 **/
double
hst_histogram_xmin(HstHistogram *hst)
{
    return hst_stats_get_min(hst->stat,0);
}

/**
 * hst_histogram_xmax:
 * @hst: a #HstHistogram
 *
 * Return value: the the upper value of the x axis
 **/
double
hst_histogram_xmax(HstHistogram *hst)
{
    return hst_stats_get_max(hst->stat,0);
}

/**
 * hst_histogram_vmin:
 * @hst: a #HstHistogram
 *
 * Return value: the minimum bin content.
 **/
double
hst_histogram_vmin(HstHistogram *hst)
{
    return hst->vmin;
}

/**
 * hst_histogram_vmax:
 * @hst: a #HstHistogram
 *
 * Return value: the maximum bin content.
 **/
double
hst_histogram_vmax(HstHistogram *hst)
{
    return hst->vmax;
}


/**
 * hst_histogram_fill:
 * @hst: a #HstHistogram
 * @v: the value
 * @w: the weight
 *
 * Increments then contents of the bin including x by w
 *
 **/
void
hst_histogram_fill(HstHistogram *hst,double v, double w)
{
    HST_HISTOGRAM_GET_CLASS(hst)->fill(hst,v,w);
    g_signal_emit (hst, hst_signals[CHANGED],0);
}

/**
 * hst_histogram_reset:
 * @hst: a #HstHistogram
 *
 * Resets to zero the contents of the histogram.
 *
 **/
void
hst_histogram_reset(HstHistogram *hst)
{
    HST_HISTOGRAM_GET_CLASS(hst)->reset(hst);
    g_signal_emit (hst, hst_signals[CHANGED],0);
}


/**
 * hst_histogram_update_stats:
 * @hst: a #HstHistogram
 *
 * Recomputes the statistics of the histogram
 *
 **/
void
hst_histogram_update_stats(HstHistogram *hst)
{
    int ix;
    HstFunction  *f;
    g_return_if_fail( HST_IS_HISTOGRAM(hst) );

    f = HST_FUNCTION(hst);
    hst_stats_reset( hst->stat );
    hst->vmin = hst->vmax = 0.;
    for (ix=0;ix<f->nbin;ix++)
    {
        double val = hst->y[ix];
        if (val!=0.)
        {
            hst_stats_add(hst->stat,hst_function_bin_center(f,ix),0.,val);
            if (val>hst->vmax)
                hst->vmax=val;
            if (val<hst->vmin)
                hst->vmin=val;
        }
    }
}



static void
hst_array(HstFunction *f, Tuple *t, double xmin, double xmax)
{
    HstHistogram *hst = HST_HISTOGRAM(f);
    HstFunction  *ff  = HST_FUNCTION(f);
    int i,j,n,i1,i2;
    double xv=0.;
    i1 = MAX(hst_function_find_bin(ff,xmin),0);
    if (i1<0)
        i1 = 0;
    if (i1==ff->nbin)
        i1 = ff->nbin-1;
    
    i2 = MIN(hst_function_find_bin(ff,xmax),ff->nbin);
    if (i2<0)
        i2 = 0;
    if (i2==ff->nbin)
        i2 = ff->nbin-1;
    
    n  = i2-i1;
    t->n = 3*n+1;
    t->P = g_new(PointList,t->n);

    for (i=i1,j=0;i<i2;i++)
    {
        t->P[j].x = f->xmin+i*f->step;
        t->P[j].y = xv;
        j++;
        t->P[j].x = t->P[j-1].x;
        t->P[j].y = hst->y[i];
        j++;
        t->P[j].x = t->P[j-1].x + f->step;
        t->P[j].y = t->P[j-1].y;
        j++;
        xv = hst->y[i];
    }
    t->P[j].x = t->P[j-1].x;
    t->P[j].y = 0.;
    j++;
    for (;j<t->n;j++)
        t->P[j].x = t->P[j].y = 0.;
    return;
}

/**
 * hst_histogram_get_stats:
 * @hst: a #HstHistogram
 * @s: a pointer to a #HstStats
 *
 * Copies into @s the contents of the @hst stats
 *
 * Return value: an integer
 **/
int
hst_histogram_get_stats(HstHistogram *hst, HstStats *s)
{
    int i;
    g_return_val_if_fail( HST_IS_HISTOGRAM(hst),-1);
    g_return_val_if_fail( s!=NULL ,-1);
    g_return_val_if_fail( HST_IS_STATS(s),-1);

    for (i=0;i<5;i++)
    {
        if (i<2)
        {
            s->xmin[i] = hst->stat->xmin[i];
            s->xmax[i] = hst->stat->xmax[i];
        }
        s->s[i] = hst->stat->s[i];
    }
    return 0;
}

/**
 * hst_histogram_get_buffer
 * @hst: a #HstHistogram
 * @sz: the address of an integer. It will contain, on exit, the length,
 *      of the buffer.
 * @buffer: the address of a double pointer. It will have on exit the
 *          address of the buffer
 *
 * Function to get the histogram buffer and its size.
 * @sz and @buffer are set to zero if problems
 **/
void
hst_histogram_get_buffer(HstHistogram *hst, int *sz, double **buffer)
{
    *sz = 0;
    *buffer = 0;
    g_return_if_fail( HST_IS_HISTOGRAM(hst) );

    g_object_get(G_OBJECT(hst),"nx",sz,"ptr",buffer,NULL);
    return;
}

static void
hst_real_changed(HstHistogram*hst) {

}
