#include "histogram2d.h"

static void   histogram2d_dispose      (GObject        *object);
static void   histogram2d_finalize     (GObject        *object);
static void   histogram2d_set_property (GObject *object, guint property_id, const GValue *value, GParamSpec *pspec);
static void   histogram2d_get_property (GObject *object, guint property_id, GValue *value, GParamSpec *pspec);
static void   histogram2d_init (HstHistogram2D * o);
static void   histogram2d_class_init (HstHistogram2DClass * c);
static void   histogram2d_fill (HstHistogram2D * h, double, double,double);
static void   histogram2d_reset (HstHistogram2D * h);
static double histogram2d_eval (HstFunction2D * f2d, double, double);
static void   histogram2d_notify_size_changed(GObject *o, GParamSpec *arg, gpointer data); 

enum {
  CHANGED,
  LAST_SIGNAL
};
enum {
    PROP_0,
    PROP_PTR,
    PROP_VMIN,
    PROP_VMAX
};

/* pointer to the class of our parent */
static HstFunction2DClass *parent_class = NULL;
static guint               hst2d_signals[LAST_SIGNAL] = { 0 };

GType
hst_histogram2d_get_type (void)
{
    static GType type = 0;

    if (type == 0) {
	static const GTypeInfo info = {
	    sizeof (HstHistogram2DClass),
	    (GBaseInitFunc) NULL,
	    (GBaseFinalizeFunc) NULL,
	    (GClassInitFunc) histogram2d_class_init,
	    (GClassFinalizeFunc) NULL,
	    NULL /* class_data */,
	    sizeof (HstHistogram2D),
	    0 /* n_preallocs */,
	    (GInstanceInitFunc) histogram2d_init,
	    NULL
	};

	type = g_type_register_static (HST_TYPE_FUNCTION2D, "HstHistogram2D", &info, (GTypeFlags)0);
    }

    return type;
}

static void 
histogram2d_init (HstHistogram2D * h2d)
{
    unsigned long npts = HST_FUNCTION2D_X(h2d)->nbin * HST_FUNCTION2D_Y(h2d)->nbin;
    h2d->vmin = 0.0;
    h2d->vmax = 0.0;
    h2d->owner=1;
    h2d->y = g_new(double,npts);
    h2d->stat= hst_stats_new();

    g_object_set_data(G_OBJECT(h2d),"__bufsiz__",(gpointer)npts);
    g_signal_connect( h2d, "notify::nx",  G_CALLBACK(histogram2d_notify_size_changed), NULL );
    g_signal_connect( h2d, "notify::ny",  G_CALLBACK(histogram2d_notify_size_changed), NULL );


}

static void 
histogram2d_class_init (HstHistogram2DClass * c)
{
    GObjectClass *g_object_class = (GObjectClass*) c;
    HstFunctionClass  *f1dclass = HST_FUNCTION_CLASS(c);
    HstFunction2DClass *f2dclass = HST_FUNCTION2D_CLASS(c);
    
    parent_class = g_type_class_ref (HST_TYPE_FUNCTION2D);

    f1dclass->array = 0;

    f2dclass->eval2d = histogram2d_eval;
    c->changed = 0;
    c->fill    = histogram2d_fill;
    c->reset   = histogram2d_reset;

    g_object_class->get_property = histogram2d_get_property;
    g_object_class->set_property = histogram2d_set_property;
    
    g_object_class->dispose = histogram2d_dispose;
    g_object_class->finalize = histogram2d_finalize;

    g_object_class_install_property (
	g_object_class,
	PROP_PTR,
	g_param_spec_pointer ("ptr", "ptr",
			      "Pointer to data",
			      G_PARAM_READABLE | G_PARAM_WRITABLE)
	);

    g_object_class_install_property (
	g_object_class,
	PROP_VMIN,
	g_param_spec_double("vmin",
			    "vmin",
			    "Minimum value of Z",
			    -DBL_MAX,DBL_MAX,0.,
			    (GParamFlags)(G_PARAM_READABLE)
	    )
	);

    g_object_class_install_property (
	g_object_class,
	PROP_VMAX,
	g_param_spec_double("vmax",
			    "vmax",
			    "Maximum value of Z",
			    -DBL_MAX,DBL_MAX,0.,
			    (GParamFlags)(G_PARAM_READABLE)
	    )
	);
    hst2d_signals[CHANGED]
	= g_signal_new ("changed",
			G_TYPE_FROM_CLASS (g_object_class),
			G_SIGNAL_RUN_FIRST,
			G_STRUCT_OFFSET (HstHistogram2DClass, changed),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
}

static void
histogram2d_set_property (GObject      *object,
			  guint         property_id,
			  const GValue *value,
			  GParamSpec   *pspec)
{
    HstHistogram2D *h;
    HstFunction2D  *f;
    void      *ptr;
    g_return_if_fail( HST_IS_HISTOGRAM2D(object) );

    h = HST_HISTOGRAM2D(object);
    f = HST_FUNCTION2D(object);
    switch (property_id) {
    case PROP_PTR:
	ptr =  g_value_get_pointer (value);
	if (ptr==0) {
	    if (h->owner)
		g_free(h->y);
	    h->y = g_new(double,HST_FUNCTION2D_X(h)->nbin*HST_FUNCTION2D_Y(h)->nbin);
	    h->owner=1;
	}
	else {
	    int ix,iy;
	    if (h->owner)
		g_free(h->y);
	    h->y = (double *)ptr;
	    h->owner=0;
	    h->vmin = 0.0;
	    h->vmax = 0.0;
	    hst_stats_reset( h->stat );
	    for (ix=0;ix< HST_FUNCTION2D_X(h)->nbin;ix++)
		for (iy=0;iy<HST_FUNCTION2D_Y(h)->nbin;iy++) {
		    int ib = hst_function2d_find_bin_from_cell(f,ix,iy);
		    double val = h->y[ib];
		    if ( val!=0. ) {
			hst_stats_add(h->stat,
				      hst_function_bin_center(HST_FUNCTION2D_X(h),ix),
				      hst_function_bin_center(HST_FUNCTION2D_Y(h),iy),
				      val
			    );
			if ( val > h->vmax)
			    h->vmax = val;
			if ( val < h->vmin )
			    h->vmin = val;
		    }
		}
	}
	break;
    default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	break;
    }
}

static void
histogram2d_get_property (GObject *object,
			  guint property_id,
			  GValue *value,
			  GParamSpec *pspec)
{
    HstHistogram2D *h2d;

    h2d = HST_HISTOGRAM2D (object);

    switch (property_id) {
    case PROP_PTR:
	g_value_set_pointer (value,(void *)h2d->y);
	break;

    case PROP_VMIN:
	g_value_set_double(value,h2d->vmin);
	break;
    case PROP_VMAX:
	g_value_set_double(value,h2d->vmax);
	break;
    default:
	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	if (&value || &pspec) break;
	break;
    }
}


static void   histogram2d_notify_size_changed(GObject *o, GParamSpec *arg, gpointer data) {
    unsigned long npts;
    unsigned long bfsiz;
    HstHistogram2D *h2d = HST_HISTOGRAM2D(o);
    if ( ! h2d->owner )
	return;
    npts = HST_FUNCTION2D_X(o)->nbin * HST_FUNCTION2D_Y(o)->nbin;
    bfsiz = (unsigned long)g_object_get_data(o,"__bufsiz__");

    if ( npts>bfsiz ) {
        h2d->y = g_realloc( h2d->y, npts*sizeof(double) );
        histogram2d_reset(h2d);
        g_object_set_data(G_OBJECT(h2d),"__bufsiz__",(gpointer)npts);
    }
}


static void   histogram2d_dispose      (GObject        *object) {
    G_OBJECT_CLASS (parent_class)->dispose (object);
}
static void   histogram2d_finalize     (GObject        *object) {
    HstHistogram2D *hst = HST_HISTOGRAM2D(object);
    
    if (hst->owner) {
	g_free(hst->y);
	hst->y=0;
    }
    g_object_unref(G_OBJECT(hst->stat));
    
    G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * hst_histogram2d_new:
 * @nx: number of bins in X axis
 * @xmin: minimum value of X
 * @xmax: maximum value of Y
 * @ny: number of bins in Y axis
 * @ymin: minimum value of Y
 * @ymax: maximum value of Y
 * @ptr: a pointer where the data will be stored. If NULL, a new one
 *       will be created.
 *
 * Creates a new HstHistogram2D
 *
 * Return value: a pointer to the new HstHistogram2D
 **/
HstHistogram2D *hst_histogram2d_new(int nx, double xmin, double xmax,
		       int ny, double ymin, double ymax, gdouble *ptr)
{
    
    HstHistogram2D *h;
    g_return_val_if_fail ( nx>0, NULL );
    g_return_val_if_fail ( ny>0, NULL );
    h = g_object_new (HST_TYPE_HISTOGRAM2D,
		      "nx",nx,"xmin",xmin,"xmax",xmax,
		      "ny",ny,"ymin",ymin,"ymax",ymax,
		      "ptr",ptr,
		      NULL);
    return h;
}

static void 
histogram2d_reset (HstHistogram2D * h2d) {
    int i,npts;
    HstFunction2D *ff;
    g_return_if_fail (h2d != NULL);
    g_return_if_fail (HST_IS_HISTOGRAM2D (h2d));

    ff = HST_FUNCTION2D(h2d);
    npts = ff->X.nbin*ff->Y->nbin;
    for (i=0;i<npts;i++)
	h2d->y[i]=0.;
}

/**
 * hst_histogram2d_fill:
 * @h2d: a HstHistogram2D
 * @x: value of x
 * @y: value of y
 * @w: weight
 *
 * This function will increase the contents of the cell corresponding
 * to @x and @y by @w. It will emmit the ::changed signal
 *
 **/
void 
hst_histogram2d_fill (HstHistogram2D * h2d, double x, double y, double w)
{
    g_return_if_fail (h2d != NULL);
    g_return_if_fail (HST_IS_HISTOGRAM2D (h2d));
    HST_HISTOGRAM2D_GET_CLASS(h2d)->fill(h2d,x,y,w);
    g_signal_emit (h2d, hst2d_signals[CHANGED],0);
}

/**
 * hst_histogram2d_reset:
 * @h2d: a HstHistogram2D
 *
 * Resets the contents of the histogram. It will emmit the ::changed
 * signal 
 *
 **/
void hst_histogram2d_reset(HstHistogram2D *h2d) {
    HST_HISTOGRAM2D_GET_CLASS(h2d)->reset(h2d);
    g_signal_emit (h2d, hst2d_signals[CHANGED],0);
}

static void 
histogram2d_fill (HstHistogram2D *h2d, double x, double y, double w) {
    int ib,ibx,iby;
    double val;
    HstFunction *fx,*fy; 
    g_return_if_fail (h2d != NULL);
    g_return_if_fail (HST_IS_HISTOGRAM2D (h2d));

    fx = HST_FUNCTION2D_X(h2d);
    fy = HST_FUNCTION2D_Y(h2d);

    ibx = hst_function_find_bin (HST_FUNCTION2D_X(h2d),x);
    if (ibx<0 || ibx>=fx->nbin)
        return;
    
    iby = hst_function_find_bin (HST_FUNCTION2D_Y(h2d),y);
    if (iby<0 || iby>=fx->nbin)
        return;
    
    ib  = hst_function2d_find_bin_from_cell(HST_FUNCTION2D(h2d),ibx,iby);
    val = (h2d->y[ ib ] += w);
    if ( val > h2d->vmax)
	h2d->vmax = val;
    if ( val < h2d->vmin )
	h2d->vmin = val;
    
    hst_stats_add(h2d->stat,x,y,w);
}

static double 
histogram2d_eval (HstFunction2D * h2d, double x, double y)
{
    int ibx,iby,ib;
    g_return_val_if_fail ( HST_IS_HISTOGRAM2D(h2d), 0. );
    ibx = hst_function_find_bin (HST_FUNCTION2D_X(h2d),x);
    if (ibx<0)
        ibx=0;
    if (ibx>=HST_FUNCTION2D_X(h2d)->nbin)
        ibx =HST_FUNCTION2D_X(h2d)->nbin-1;
    
    iby = hst_function_find_bin (HST_FUNCTION2D_Y(h2d),y);
    if (iby<0)
        iby=0;
    if (iby>=HST_FUNCTION2D_Y(h2d)->nbin)
        iby =HST_FUNCTION2D_Y(h2d)->nbin-1;
    
    ib  = hst_function2d_find_bin_from_cell(HST_FUNCTION2D(h2d),ibx,iby);
    return HST_HISTOGRAM2D(h2d)->y[ib];
}


/**
 * hst_histogram2d_get_stats
 * @h2d: a #HstHistogram2D
 * @s: a pointer to a #HstStats
 *
 * Copies into @s the contents of the @hst stats
 *
 * Return value: an integer
 **/
int hst_histogram2d_get_stats(HstHistogram2D *h2d, HstStats *s) {
    int i;
    g_return_val_if_fail( HST_IS_HISTOGRAM2D(h2d),-1);
    g_return_val_if_fail( s!=NULL ,-1);
    g_return_val_if_fail( HST_IS_STATS(s),-1);

    s->nc = h2d->stat->nc; 
    for (i=0;i<5;i++) {
	if (i<2) {
	    s->xmin[i] = h2d->stat->xmin[i];
	    s->xmax[i] = h2d->stat->xmax[i];
	}
	s->s[i] = h2d->stat->s[i];
    }
    
    return 0;
}

